/**
   This class provides the Base 64 encoding for hashed password.

   Solves CS166 homework assignment #3
   @author SJSU Students
   @version 1.01 10/24/2012
 */

public class Base64
{
   /**
      Method will pad necessary zeroes on the byte
      @param length length of the byte array.
      @param bytes the byte array that will be padded with zeroes
      @return byte value for the padded zeroes
    */
   public static byte[] zeroPad(int length, byte[] bytes)
   {
      byte[] padded = new byte[length]; // initialized to zero by JVM
      System.arraycopy(bytes, 0, padded, 0, bytes.length);
      return padded;
   }

   /**
      This method encodes the argument that is passed in with base64 encoding.
      @param passwordBytes the bytes that will be encoded with base64 encoding
      @return the encoded base64 encoding string.
    */
   public static String encode(byte[] passwordBytes)
   {

      String encoded = "";
      byte[] stringArray = passwordBytes;

        // determine how many padding bytes to add to the output
      int paddingCount = (3 - (stringArray.length % 3)) % 3;
      // add any necessary padding to the input
      stringArray = zeroPad(stringArray.length + paddingCount, stringArray);
      // process 3 bytes at a time, churning out 4 output bytes
      // worry about CRLF insertions later
      for (int i = 0; i < stringArray.length; i += 3) {
         int j = ((stringArray[i] & 0xff) << 16)
                 + ((stringArray[i + 1] & 0xff) << 8)
                 + (stringArray[i + 2] & 0xff);
         encoded = encoded + base64code.charAt((j >> 18) & 0x3f)
                 + base64code.charAt((j >> 12) & 0x3f)
                 + base64code.charAt((j >> 6) & 0x3f)
                 + base64code.charAt(j & 0x3f);
      }
      // replace encoded padding nulls with "="
      return splitLines(encoded.substring(0, encoded.length()
              - paddingCount) + "==".substring(0, paddingCount));

   }

   /**
      This method will split the lines and pad any null values with = at the end
      @param string the string that will be split and padded with =
      @return String that will be returned after padding, this is the true
              base64 encoding string.
    */
   public static String splitLines(String string)
   {
      String lines = "";
      for (int i = 0; i < string.length(); i += splitLinesAt)
      {
         lines += string.substring(i, Math.min
                 (string.length(), i + splitLinesAt));
         lines += "\r\n";
      }
      return lines;
   }

   private static final String base64code = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
           + "abcdefghijklmnopqrstuvwxyz" + "0123456789" + "+/";
   private static final int splitLinesAt = 76;
}
