import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;

/**
 This class is the password tool for changing password with or without salt and
 checking password, it also cracks passwords if provided with a dictionary file.

 
 @author  SJSU Students
 @version 1.01 10/24/2012
 */
public class PasswordTool
{
   /**
    This is the main class that will run the password tool

    @param args the arguments that are passed in: args[0] is the operation, such
    as change_no_salt, change_salt, crack, check. args[1] is the total number of
    columns in the password line. args[2] is the column number where the
    username is (this is usually 0). args[3] is the column number where the
    password resides. args[4] is the username (args[4] for crack is the
    dictionary file). args[5] is the raw password. (args[5] for crack is the
    password file). args[6] is the password file for the action check,
    change_no_salt, or change_salt.
    */
   public static void main(String[] args) throws NoSuchAlgorithmException
   {
      PasswordTool tool = new PasswordTool();
      tool.operation = args[0];
      tool.numberColumns = Integer.parseInt(args[1]);
      tool.loginColumnNumber = Integer.parseInt(args[2]);
      tool.passwordColumnNumber = Integer.parseInt(args[3]);
      tool.login = args[4];
      tool.password = args[5];
      tool.checkInput(tool, args);
   }

   /**
    This method checks the user's argument input and decides what action to take

    @param tool The instance of the password tool
    @param args the arguments that users put in to the command line argument
    */
   public void checkInput(PasswordTool tool, String[] args)
           throws NoSuchAlgorithmException
   {
      if (tool.operation.equalsIgnoreCase("change_salt"))
      {
         ArrayList<String> myPasswordList = tool.readFile(args[6]);
         tool.changeSalt(myPasswordList, args[6]);
      }
      else if (tool.operation.equalsIgnoreCase("change_no_salt"))
      {
         ArrayList<String> myPasswordList = tool.readFile(args[6]);
         tool.changeNoSalt(myPasswordList, args[6]);
      }
      else if (tool.operation.equalsIgnoreCase("check"))
      {
         ArrayList<String> myPasswordList = tool.readFile(args[6]);
         System.out.println(tool.check(myPasswordList, tool.password));
      }
      else if (tool.operation.equalsIgnoreCase("crack"))
      {
         ArrayList<String> dictionaryList = tool.readFile(args[4]);
         ArrayList<String> myPasswordList = tool.readFile(args[5]);
         tool.crack(dictionaryList, myPasswordList);
      }
   }

   /**
    This method changes a password with no salt

    @param myPasswordList the array-list containing the passwords from password
    file
    @param filename the filename where salt will be changed in.
    */
   public void changeNoSalt(ArrayList<String> myPasswordList, String filename)
   {
      boolean added = false;
      try
      {
         for (int x = 0; x < myPasswordList.size() && !added; x++)
         {
            added = changeNoSaltLoginFound(myPasswordList, x, added);
         }
         if (!added)
         {
            myPasswordList.add(newPassword());
         }
         output(myPasswordList, filename); //save the changes to password file
      }
      catch (Exception ex)
      {
         System.out.println(ex.getMessage());
      }
   }

   /**
    This method changes the password with salt

    @param myPasswordList the array-list containing the passwords from password
    file
    @param filename the name of the stored password file
    */
   public void changeSalt(ArrayList<String> myPasswordList, String filename)
   {
      boolean added = false;
      try
      {
         for (int x = 0; x < myPasswordList.size() && !added; x++)
         {
            added = changeSaltLoginFoundNotFound(myPasswordList, x, added);
         }
         if (!added)
         {
            myPasswordList.add(newSaltedPassword());
         }
         output(myPasswordList, filename);  //save changes to password file
      }//end try
      catch (Exception ex)
      {
         System.out.println(ex.getMessage());
      }
   }

   /**
    This method cracks the password

    @param dictionaryList the arrayList containing the list of passwords.
    @param myList the array-list containing the passwords and usernames.
    */
   public void crack(ArrayList<String> dictionaryList,
                     ArrayList<String> myList) throws NoSuchAlgorithmException
   {
      String dictionaryPassword;
      for (int i = 0; i < myList.size(); i++)
      {
         String[] fields = myList.get(i).split(":");
         try
         {
            String[] saltAndPassword = fields[passwordColumnNumber].
                    substring(1).split("\\$");  //extract password field

            for (int j = 0; j < dictionaryList.size(); j++)
            {
               dictionaryPassword = crackSaltAndNoSalt(saltAndPassword,
                       dictionaryList, j);
               printCrackedPassword(dictionaryPassword,
                       saltAndPassword, fields, dictionaryList, j);
            }
         }
         catch (IndexOutOfBoundsException ex)
         {
            System.out.println(ex.getMessage());
         }
      }
   }

   /**
    This method is used to output the changes to the password file

    @param myPasswordList the array-list containing the passwords to output to
    the file
    @param filename the filename to output changes to
    */
   public void output(ArrayList<String> myPasswordList, String filename)
   {
      FileWriter fileStream;
      try
      {
         fileStream = new FileWriter(filename);
         BufferedWriter out = new BufferedWriter(fileStream);
         for (int z = 0; z < myPasswordList.size(); z++)
         {
            out.write(myPasswordList.get(z));
            out.newLine();
         }
         out.close();
      }
      catch (IOException ex)
      {
         System.out.println(ex);
      }
   }

   /**
    This method creates a new password string for a user only the login and
    password columns have a value, no salt is used

    @return the new password string
    */
   public String newPassword() throws NoSuchAlgorithmException
   {
      MessageDigest md = MessageDigest.getInstance("MD5");
      md.update(password.getBytes(), 0, password.length());
      byte[] md5 = md.digest();
      String aString = "";
      for (int occurance = 0, x = 0; x < numberColumns; x++)
      {
         aString = checkInsertionColUnsalted(x, occurance, aString, md5);
      }
      return aString;
   }

   /**
    This class will generate a new salted password for usernames that are not in
    the password file.

    @return the password String of the new salted password.
    */
   public String newSaltedPassword() throws NoSuchAlgorithmException
   {
      int salt = getSalt();
      password = password + Integer.toString(salt);
      MessageDigest md = MessageDigest.getInstance("MD5");
      md.update(password.getBytes(), 0, password.length());
      byte[] md5 = md.digest();
      String aString = "";
      for (int occurance = 0, x = 0; x < numberColumns; x++)
      {
         aString = checkInsertionColumnSalted(x, occurance, aString, salt, md5);
      }
      return aString;
   }

   /**
    This method checks if a login exists if login & password match prints fail
    if login found, but password didn't match prints not found if login is not
    found

    @param myPasswordList the array-list containing the password strings prints
    pass
    @param rawPassword the password to check with
    @return String that says PASS, FAIL, or NOT_FOUND. PASS if the username and
    password matches. FAIL if the username is correct, but password does not
    match. NOT_FOUND is the username is not in the file.
    */
   public String check(ArrayList<String> myPasswordList, String rawPassword)
   {
      try
      {
         for (int x = 0; x < myPasswordList.size(); x++)
         {
            String[] fields = myPasswordList.get(x).split(":");
            try
            {
               if ((fields[loginColumnNumber].equals(login)))  //login found
               {
                  return checkLoginFound(fields, rawPassword);
               }
               else if (x == myPasswordList.size() - 1)  //login not found
               {
                  return "NOT_FOUND";
               }
            }
            catch (IndexOutOfBoundsException ex)
            {
            }
         }
      }
      catch (Exception ex)
      {
         return ex.getMessage();
      }
      return null;
   }

   /**
    This method is used to get a random salt

    @return random salt value
    */
   public int getSalt()
   {
      return (int) (Math.random() * Integer.MAX_VALUE);
   }

   /**
    This method is used to read the password file line by line

    @param filename the name of the password file
    @return an array-list containing the password lines
    */
   public ArrayList<String> readFile(String filename)
   {
      ArrayList<String> textContent = new ArrayList<String>();
      try
      {
         for (String line : new TextFileReader(filename))
         {
            textContent.add(line);
         }
      }
      catch (Exception e)
      {
         //Catch exception if any
         System.err.println("Error: " + e.getMessage());
      }
      return textContent;
   }

   public boolean loginFoundChangeNoSalt(String[] fields,
           ArrayList<String> myPasswordList, int x)
           throws NoSuchAlgorithmException
   {
      boolean added;
      MessageDigest md = MessageDigest.getInstance("MD5");
      md.update(password.getBytes(), 0, password.length());
      byte[] md5 = md.digest();
      String encoded64 = "$0$" + Base64.encode(md5).trim();
      fields[passwordColumnNumber] = encoded64;
      int occurance = 0;
      StringBuilder s = new StringBuilder();
      for (int y = 0; y < fields.length; y++)
      {
         if (y == fields.length - 1)
         {
            s.append(fields[y]);
         }
         else if (y < fields.length - 1)
         {
            s.append(fields[y]).append(":");
            occurance++;
         }
      }
      String newString = s.toString();
      while (occurance < numberColumns - 1)
      {
         newString = newString + ":";
         occurance++;
      }
      myPasswordList.set(x, newString);
      added = true;
      return added;
   }

   /**
    This method will change the passwords with salt when the login is found in
    change_salt action.

    @param fields the array containing the fields in the password file, such as
    the usernames, salt or no salt, passwords, etc
    @param myPasswordList the password list from password file.
    @param size the size of the password list
    @return boolean that indicates whether password has been successfully
    replaced.
    */
   public boolean loginFoundChangeSalt(String[] fields,
           ArrayList<String> myPasswordList, int size)
           throws NoSuchAlgorithmException
   {
      hashAndEncodePasswordWithSalt(fields);
      int occurance = 0;
      StringBuilder s = new StringBuilder();
      for (int y = 0; y < fields.length; y++)  //build new password string
      {
         if (y == fields.length - 1)
         {
            s.append(fields[y]);
         }
         else if (y < fields.length - 1)
         {
            s.append(fields[y]).append(":");
            occurance++;
         }
      }
      String newString = s.toString();
      while (occurance < numberColumns - 1)
      {
         newString = newString + ":";
         occurance++;
      }
      myPasswordList.set(size, newString);  //replace old password string
      //with new built password string
      boolean added = true;
      return added;
   }

   /**
    This method will crack the salted and non-salted password that is in the
    password and dictionary files.

    @param saltAndPassword the array list containing all the salt and passwords
    in the password file.
    @param dictionaryList the list of passwords in the dictionary
    @param j the size of the dictionary list.
    @return the password strings that are cracked.
    @throws NumberFormatException
    */
   public String crackSaltAndNoSalt(String[] saltAndPassword,
                                    ArrayList<String> dictionaryList, int j)
           throws NumberFormatException
   {
      String dictionaryPassword;
      if (Integer.parseInt(saltAndPassword[0]) > 0)
      {
         //salt
         dictionaryPassword = dictionaryList.get(j);
         dictionaryPassword = dictionaryPassword + saltAndPassword[1];
      }
      else
      {
         //no salt
         dictionaryPassword = dictionaryList.get(j);
      }
      return dictionaryPassword;
   }

   /**
    This method will print out the username and cracked password.

    @param dictionaryPassword the raw passwords that have been cracked.
    @param saltAndPassword the array list containing salt and passwords from the
    passwords file.
    @param fields The string of fields containing the username information.
    @param dictionaryList the list containing the password from dictionary.txt
    @param j the size of the dictionary
    */
   public void printCrackedPassword(String dictionaryPassword,
                                    String[] saltAndPassword, String[] fields,
                                    ArrayList<String> dictionaryList, int j)
           throws NoSuchAlgorithmException
   {
      boolean found;
      MessageDigest md = MessageDigest.getInstance("MD5");
      md.update(dictionaryPassword.getBytes(), 0, dictionaryPassword.length());
      byte[] md5 = md.digest();
      String encoded64 = Base64.encode(md5).trim();
      if (encoded64.equals(saltAndPassword[saltAndPassword.length - 1]))
      {
         System.out.println(fields[loginColumnNumber]
                 + " " + dictionaryList.get(j));
         found = true;
      }
   }

   /**
    This method will check if the passwords match if the login is found.

    @param fields the fields containing all the passwords from the password
    file.
    @param rawPassword raw password that is passed in to be checked
    @return String containing PASS or FAIL message PASS if the password matches,
    FAIL otherwise
    @throws NumberFormatException
    */
   public String checkLoginFound(String[] fields, String rawPassword)
           throws NumberFormatException, NoSuchAlgorithmException
   {
      String[] saltAndPassword = fields[passwordColumnNumber].
              substring(1).split("\\$");  //extract password field

      if (Integer.parseInt(saltAndPassword[0]) > 0)
      {
         rawPassword = rawPassword + saltAndPassword[1];
      }

      MessageDigest md = MessageDigest.getInstance("MD5");
      md.update(rawPassword.getBytes(), 0, rawPassword.length());
      byte[] md5 = md.digest();

      if (Base64.encode(md5).trim().
              equals(saltAndPassword[saltAndPassword.length - 1]))
      {
         return "PASS";
      }
      else
      {
         return "FAIL";
      }
   }

   /**
    This method will hash and encode passwords with salt

    @param fields the array containing the fields in the password file, such as
    the usernames, salt or no salt, passwords, etc
    */
   public void hashAndEncodePasswordWithSalt(String[] fields)
           throws NoSuchAlgorithmException
   {
      int salt = getSalt();  //get random salt
      //append salt password before hashing
      password = password + String.valueOf(salt);
      MessageDigest md = MessageDigest.getInstance("MD5");
      md.update(password.getBytes(), 0, password.length());
      byte[] md5 = md.digest();
      //base64Encode(md5) and create password string base64Encoded
      String encoded64 = "$1$" + String.valueOf(
              salt) + "$" + Base64.encode(md5).trim();
      //replace previous password field with new password string
      fields[passwordColumnNumber] = encoded64;
   }

   /**
   This method assist the changeNoSalt method if the login is found or not found

   @param myPasswordList the list containing the passwords
   @param x the counter in the for loop
   @param added boolean that is returned if password is added.
   @return boolean that determines if the password is added
   @throws NoSuchAlgorithmException
   */
   public boolean changeNoSaltLoginFound(ArrayList<String> myPasswordList,
                   int x, boolean added) throws NoSuchAlgorithmException
   {
      String[] fields = myPasswordList.get(x).split(":");
      try
      {
         if ((fields[loginColumnNumber].equals(login))) //login found
         {
            added = loginFoundChangeNoSalt(fields, myPasswordList, x);
         }
         //the login is not found
         else if (x == myPasswordList.size() - 1 && !added)
         {
            myPasswordList.add(newPassword()); //new password string
            added = true;
         }
      }
      catch (IndexOutOfBoundsException ex)
      {
         System.out.println(ex.getMessage());
      }
      return added;
   }

   /**
   This method assist the change salt method if the login is found or not found.

   @param myPasswordList the list of password
   @param x the for loop counter
   @param added boolean that is returned if password is added
   @return boolean that determines if the password is added
   @throws NoSuchAlgorithmException
   */
   public boolean changeSaltLoginFoundNotFound(ArrayList<String> myPasswordList,
                          int x, boolean added) throws NoSuchAlgorithmException
   {
      String[] fields = myPasswordList.get(x).split(":");
      try
      {
         if ((fields[loginColumnNumber].equals(login)))  //login found
         {
            added = loginFoundChangeSalt(fields, myPasswordList, x);
         }
         else if (x == myPasswordList.size() - 1 && !added) //not found
         {
            myPasswordList.add(newSaltedPassword()); //new password string
            added = true;
         }
      }
      catch (IndexOutOfBoundsException ex)
      {
         System.out.println(ex.getMessage());
      }
      return added;
   }

   /**
   This is a helper method for newSaltedPassword method.
   This method will check in which column the password and column will be
   inserted and append : as appropriate for proper formatting
   @param x the counter for the for loop that this method resides in
   @param occurance the number of occurences for colons
   @param aString the password and login string that will be printed out
   @param md5 the md5 hash
   @return string that will be printed out to the file
   */
   public String checkInsertionColumnSalted(int x, int occurance,
                                           String aString, int salt, byte[] md5)
   {
      if (loginColumnNumber == x && occurance < numberColumns - 1)
      {
         aString = aString + login + ":";
         occurance++;
      }
      else if (passwordColumnNumber == x && occurance < numberColumns - 1)
      {
         aString = aString + "$1$" + Integer.toString(salt)
                 + "$" + Base64.encode(md5).trim() + ":";
         occurance++;
      }
      else if (passwordColumnNumber == x && x == numberColumns - 1)
      {
         aString = aString + "$1$" + Integer.toString(salt)
                 + "$" + Base64.encode(md5).trim();
      }
      else if (loginColumnNumber == x && x == numberColumns - 1)
      {
         aString = aString + login;
      }
      else if (occurance < numberColumns - 1)
      {
         aString = aString + ":";
         occurance++;
      }
      return aString;
   }

   /**
   This is a helper method for newPassword method.
   This method will check in which column the password and column will be
   inserted and append : as appropriate for proper formatting
   @param x the counter for the for loop that this method resides in
   @param occurance the number of occurences for colons
   @param aString the password and login string that will be printed out
   @param md5 the md5 hash
   @return string that will be printed out to the file
   */
   public String checkInsertionColUnsalted(int x, int occurance, String aString,
                                                                 byte[] md5)
   {
      if (loginColumnNumber == x && occurance < numberColumns - 1)
      {
         aString = aString + login + ":";
         occurance++;
      }
      else if (passwordColumnNumber == x && occurance < numberColumns - 1)
      {
         aString = aString + "$0$" + Base64.encode(md5).trim() + ":";
         occurance++;
      }
      else if (passwordColumnNumber == x && x == numberColumns - 1)
      {
         aString = aString + "$0$" + Base64.encode(md5).trim();
      }
      else if (loginColumnNumber == x && x == numberColumns - 1)
      {
         aString = aString + login;
      }
      else if (occurance < numberColumns - 1)
      {
         aString = aString + ":";
         occurance++;
      }
      return aString;
   }
   private int numberColumns;
   private int loginColumnNumber;
   private int passwordColumnNumber;
   private String login;
   private String operation;
   private String password;
}
